@@ -18,13 +18,14 @@
 
 #include "ntpd.h"
 
+u_int constraint_cnt = 0;
+
 int
 constraint_init(struct constraint *cstr)
 {
 	return (0);
 }
 
-
 int
 constraint_query(struct constraint *cstr)
 {
@@ -32,21 +33,52 @@ constraint_query(struct constraint *cstr)
 }
 
 void
-constraint_check_child(void)
+priv_constraint_check_child(pid_t pid, int status)
+{
+}
+
+void
+priv_constraint_kill(u_int32_t id)
+{
+}
+
+void
+priv_constraint_msg(u_int32_t id, u_int8_t *data, size_t len,
+    const char *pw_dir, uid_t pw_uid, gid_t pw_gid)
+{
+}
+
+void
+constraint_purge(void)
+{
+}
+
+void
+constraint_add(struct constraint *cstr)
+{
+}
+
+void
+constraint_msg_dns(u_int32_t id, u_int8_t *data, size_t len)
+{
+}
+
+void
+constraint_msg_result(u_int32_t id, u_int8_t *data, size_t len)
+{
+}
+
+void
+constraint_msg_close(u_int32_t id, u_int8_t *data, size_t len)
 {
 }
 
 int
-constraint_dispatch_msg(struct pollfd *pfd)
+priv_constraint_dispatch(struct pollfd *pfd)
 {
 	return (0);
 }
 
-void
-constraint_dns(u_int32_t id, u_int8_t *data, size_t len)
-{
-}
-
 int
 constraint_check(double val)
 {
