@@ -1,4 +1,4 @@
-/*	$OpenBSD: ntp.c,v 1.129 2015/02/12 01:54:57 reyk Exp $ */
+/*	$OpenBSD: ntp.c,v 1.140 2015/12/05 13:12:16 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -28,36 +28,32 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
-#include <tls.h>
+#include <err.h>
 
 #include "ntpd.h"
 
 #define	PFD_PIPE_MAIN	0
-#define	PFD_HOTPLUG	1
-#define	PFD_PIPE_DNS	2
-#define	PFD_SOCK_CTL	3
-#define	PFD_MAX		4
+#define	PFD_PIPE_DNS	1
+#define	PFD_SOCK_CTL	2
+#define	PFD_MAX		3
 
 volatile sig_atomic_t	 ntp_quit = 0;
-volatile sig_atomic_t	 ntp_report = 0;
-volatile sig_atomic_t	 ntp_sigchld = 0;
 struct imsgbuf		*ibuf_main;
 struct imsgbuf		*ibuf_dns;
 struct ntpd_conf	*conf;
 struct ctl_conns	 ctl_conns;
 u_int			 peer_cnt;
 u_int			 sensors_cnt;
-u_int			 constraint_cnt;
-time_t			 lastreport;
+extern u_int		 constraint_cnt;
 
 void	ntp_sighdlr(int);
 int	ntp_dispatch_imsg(void);
 int	ntp_dispatch_imsg_dns(void);
 void	peer_add(struct ntp_peer *);
 void	peer_remove(struct ntp_peer *);
-void	report_peers(int);
 
 void
 ntp_sighdlr(int sig)
@@ -67,12 +63,6 @@ ntp_sighdlr(int sig)
 	case SIGTERM:
 		ntp_quit = 1;
 		break;
-	case SIGINFO:
-		ntp_report = 1;
-		break;
-	case SIGCHLD:
-		ntp_sigchld = 1;
-		break;
 	}
 }
 
@@ -81,7 +71,7 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
     struct passwd *pw)
 {
 	int			 a, b, nfds, i, j, idx_peers, timeout;
-	int			 hotplugfd, nullfd, pipe_dns[2], idx_clients;
+	int			 nullfd, pipe_dns[2], idx_clients;
 	int			 ctls;
 	u_int			 pfd_elms = 0, idx2peer_elms = 0;
 	u_int			 listener_cnt, new_cnt, sent_cnt, trial_cnt;
@@ -110,25 +100,19 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 		return (pid);
 	}
 
-	tls_init();
-
-	/* Verification will be turned off if CA is not found */
-	if ((conf->ca = tls_load_file(CONSTRAINT_CA,
-	    &conf->ca_len, NULL)) == NULL)
-		log_warnx("constraint certificate verification turned off");
-
 	/* in this case the parent didn't init logging and didn't daemonize */
 	if (nconf->settime && !nconf->debug) {
-		log_init(nconf->debug);
+		log_init(nconf->debug, LOG_DAEMON);
 		if (setsid() == -1)
 			fatal("setsid");
 	}
+	log_procinit("ntp");
+
 	if ((se = getservbyname("ntp", "udp")) == NULL)
 		fatal("getservbyname");
 
-	if ((nullfd = open(_PATH_DEVNULL, O_RDWR, 0)) == -1)
+	if ((nullfd = open("/dev/null", O_RDWR, 0)) == -1)
 		fatal(NULL);
-	hotplugfd = sensor_hotplugfd();
 
 	close(pipe_prnt[0]);
 	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe_dns) == -1)
@@ -167,12 +151,15 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 
 	endservent();
 
+	/* The ntp process will want to open NTP client sockets -> "inet" */
+	if (pledge("stdio inet", NULL) == -1)
+		err(1, "pledge");
+
 	signal(SIGTERM, ntp_sighdlr);
 	signal(SIGINT, ntp_sighdlr);
-	signal(SIGINFO, ntp_sighdlr);
 	signal(SIGPIPE, SIG_IGN);
 	signal(SIGHUP, SIG_IGN);
-	signal(SIGCHLD, ntp_sighdlr);
+	signal(SIGCHLD, SIG_DFL);
 
 	if ((ibuf_main = malloc(sizeof(struct imsgbuf))) == NULL)
 		fatal(NULL);
@@ -190,7 +177,7 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 	TAILQ_FOREACH(p, &conf->ntp_peers, entry)
 		client_peer_init(p);
 
-	bzero(&conf->status, sizeof(conf->status));
+	memset(&conf->status, 0, sizeof(conf->status));
 
 	conf->freq.num = 0;
 	conf->freq.samples = 0;
@@ -218,9 +205,6 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 	TAILQ_FOREACH(p, &conf->ntp_peers, entry)
 		peer_cnt++;
 
-	/* wait 5 min before reporting first status to let things settle down */
-	lastreport = getmonotime() + (5 * 60) - REPORT_INTERVAL;
-
 	while (ntp_quit == 0) {
 		if (peer_cnt > idx2peer_elms) {
 			if ((newp = reallocarray(idx2peer, peer_cnt,
@@ -235,7 +219,7 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 		}
 
 		new_cnt = PFD_MAX +
-		    peer_cnt + listener_cnt + ctl_cnt + constraint_cnt;
+		    peer_cnt + listener_cnt + ctl_cnt;
 		if (new_cnt > pfd_elms) {
 			if ((newp = reallocarray(pfd, new_cnt,
 			    sizeof(*pfd))) == NULL) {
@@ -248,13 +232,11 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 			pfd_elms = new_cnt;
 		}
 
-		bzero(pfd, sizeof(*pfd) * pfd_elms);
-		bzero(idx2peer, sizeof(*idx2peer) * idx2peer_elms);
+		memset(pfd, 0, sizeof(*pfd) * pfd_elms);
+		memset(idx2peer, 0, sizeof(*idx2peer) * idx2peer_elms);
 		nextaction = getmonotime() + 3600;
 		pfd[PFD_PIPE_MAIN].fd = ibuf_main->fd;
 		pfd[PFD_PIPE_MAIN].events = POLLIN;
-		pfd[PFD_HOTPLUG].fd = hotplugfd;
-		pfd[PFD_HOTPLUG].events = POLLIN;
 		pfd[PFD_PIPE_DNS].fd = ibuf_dns->fd;
 		pfd[PFD_PIPE_DNS].events = POLLIN;
 		pfd[PFD_SOCK_CTL].fd = fd_ctl;
@@ -282,9 +264,8 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 			if (p->deadline > 0 && p->deadline <= getmonotime()) {
 				timeout = 300;
 				log_debug("no reply from %s received in time, "
-				    "next query %ds %s", log_sockaddr(
-				    (struct sockaddr *)&p->addr->ss), timeout,
-				    print_rtable(p->rtable));
+				    "next query %ds", log_sockaddr(
+				    (struct sockaddr *)&p->addr->ss), timeout);
 				if (p->trustlevel >= TRUSTLEVEL_BADPEER &&
 				    (p->trustlevel /= 2) < TRUSTLEVEL_BADPEER)
 					log_info("peer %s now invalid",
@@ -317,21 +298,23 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 		}
 		idx_clients = i;
 
-		if (last_sensor_scan == 0 ||
-		    last_sensor_scan + SENSOR_SCAN_INTERVAL < getmonotime()) {
-			sensors_cnt = sensor_scan();
-			last_sensor_scan = getmonotime();
-		}
-		if (!TAILQ_EMPTY(&conf->ntp_conf_sensors) && sensors_cnt == 0 &&
-		    nextaction > last_sensor_scan + SENSOR_SCAN_INTERVAL)
-			nextaction = last_sensor_scan + SENSOR_SCAN_INTERVAL;
-		sensors_cnt = 0;
-		TAILQ_FOREACH(s, &conf->ntp_sensors, entry) {
-			if (conf->settime && s->offsets[0].offset)
-				priv_settime(s->offsets[0].offset);
-			sensors_cnt++;
-			if (s->next > 0 && s->next < nextaction)
-				nextaction = s->next;
+		if (!TAILQ_EMPTY(&conf->ntp_conf_sensors)) {
+			if (last_sensor_scan == 0 ||
+			    last_sensor_scan + SENSOR_SCAN_INTERVAL <= getmonotime()) {
+				sensors_cnt = sensor_scan();
+				last_sensor_scan = getmonotime();
+			}
+			if (sensors_cnt == 0 &&
+			    nextaction > last_sensor_scan + SENSOR_SCAN_INTERVAL)
+				nextaction = last_sensor_scan + SENSOR_SCAN_INTERVAL;
+			sensors_cnt = 0;
+			TAILQ_FOREACH(s, &conf->ntp_sensors, entry) {
+				if (conf->settime && s->offsets[0].offset)
+					priv_settime(s->offsets[0].offset);
+				sensors_cnt++;
+				if (s->next > 0 && s->next < nextaction)
+					nextaction = s->next;
+			}
 		}
 
 		if (conf->settime &&
@@ -356,9 +339,6 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 		TAILQ_FOREACH(cstr, &conf->constraints, entry) {
 			if (constraint_query(cstr) == -1)
 				continue;
-			pfd[i].fd = cstr->fd;
-			pfd[i].events = POLLIN;
-			i++;
 		}
 
 		now = getmonotime();
@@ -406,11 +386,6 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 			ctl_cnt += control_accept(fd_ctl);
 		}
 
-		if (nfds > 0 && pfd[PFD_HOTPLUG].revents & (POLLIN|POLLERR)) {
-			nfds--;
-			sensor_hotplugevent(hotplugfd);
-		}
-
 		for (j = PFD_MAX; nfds > 0 && j < idx_peers; j++)
 			if (pfd[j].revents & (POLLIN|POLLERR)) {
 				nfds--;
@@ -431,23 +406,12 @@ ntp_main(int pipe_prnt[2], int fd_ctl, struct ntpd_conf *nconf,
 			nfds -= control_dispatch_msg(&pfd[j], &ctl_cnt);
 		}
 
-		for (; nfds > 0 && j < i; j++) {
-			nfds -= constraint_dispatch_msg(&pfd[j]);
-		}
-
 		for (s = TAILQ_FIRST(&conf->ntp_sensors); s != NULL;
 		    s = next_s) {
 			next_s = TAILQ_NEXT(s, entry);
 			if (s->next <= getmonotime())
 				sensor_query(s);
 		}
-		report_peers(ntp_report);
-		ntp_report = 0;
-
-		if (ntp_sigchld) {
-			constraint_check_child();
-			ntp_sigchld = 0;
-		}
 	}
 
 	msgbuf_write(&ibuf_main->w);
@@ -467,7 +431,7 @@ ntp_dispatch_imsg(void)
 	struct imsg		 imsg;
 	int			 n;
 
-	if ((n = imsg_read(ibuf_main)) == -1)
+	if ((n = imsg_read(ibuf_main)) == -1 && errno != EAGAIN)
 		return (-1);
 
 	if (n == 0) {	/* connection closed */
@@ -493,6 +457,14 @@ ntp_dispatch_imsg(void)
 				conf->status.synced = 0;
 			}
 			break;
+		case IMSG_CONSTRAINT_RESULT:
+			constraint_msg_result(imsg.hdr.peerid,
+			    imsg.data, imsg.hdr.len - IMSG_HEADER_SIZE);
+			break;
+		case IMSG_CONSTRAINT_CLOSE:
+			constraint_msg_close(imsg.hdr.peerid,
+			    imsg.data, imsg.hdr.len - IMSG_HEADER_SIZE);
+			break;
 		default:
 			break;
 		}
@@ -563,7 +535,6 @@ ntp_dispatch_imsg_dns(void)
 					npeer->addr_head.name =
 					    peer->addr_head.name;
 					npeer->addr_head.pool = 1;
-					npeer->rtable = peer->rtable;
 					client_peer_init(npeer);
 					npeer->state = STATE_DNS_DONE;
 					peer_add(npeer);
@@ -582,7 +553,7 @@ ntp_dispatch_imsg_dns(void)
 				client_addr_init(peer);
 			break;
 		case IMSG_CONSTRAINT_DNS:
-			constraint_dns(imsg.hdr.peerid,
+			constraint_msg_dns(imsg.hdr.peerid,
 			    imsg.data, imsg.hdr.len - IMSG_HEADER_SIZE);
 			break;
 		default:
@@ -808,60 +779,3 @@ error_interval(void)
 	r = arc4random_uniform(interval / 10);
 	return (interval + r);
 }
-
-void
-report_peers(int always)
-{
-	time_t now;
-	u_int badpeers = 0;
-	u_int badsensors = 0;
-	struct ntp_peer *p;
-	struct ntp_sensor *s;
-
-	TAILQ_FOREACH(p, &conf->ntp_peers, entry) {
-		if (p->trustlevel < TRUSTLEVEL_BADPEER)
-			badpeers++;
-	}
-	TAILQ_FOREACH(s, &conf->ntp_sensors, entry) {
-		if (!s->update.good)
-			badsensors++;
-	}
-
-	now = getmonotime();
-	if (!always) {
-		if ((peer_cnt == 0 || badpeers == 0 || badpeers < peer_cnt / 2)
-		    && (sensors_cnt == 0 || badsensors == 0 ||
-		    badsensors < sensors_cnt / 2))
-			return;
-
-		if (lastreport + REPORT_INTERVAL > now)
-			return;
-	}
-	lastreport = now;
-	if (peer_cnt > 0) {
-		log_warnx("%u out of %u peers valid", peer_cnt - badpeers,
-		    peer_cnt);
-		TAILQ_FOREACH(p, &conf->ntp_peers, entry) {
-			if (p->trustlevel < TRUSTLEVEL_BADPEER) {
-				const char *a = "not resolved";
-				const char *pool = "";
-				if (p->addr)
-					a = log_sockaddr(
-					    (struct sockaddr *)&p->addr->ss);
-				if (p->addr_head.pool)
-					pool = "from pool ";
-				log_warnx("bad peer %s%s (%s) %s",
-				    pool, p->addr_head.name, a,
-				    print_rtable(p->rtable));
-			}
-		}
-	}
-	if (sensors_cnt > 0) {
-		log_warnx("%u out of %u sensors valid",
-		    sensors_cnt - badsensors, sensors_cnt);
-		TAILQ_FOREACH(s, &conf->ntp_sensors, entry) {
-			if (!s->update.good)
-				log_warnx("bad sensor %s", s->device);
-		}
-	}
-}
