@@ -1,4 +1,4 @@
-/*	$OpenBSD: ntpd.c,v 1.92 2015/02/11 03:16:57 reyk Exp $ */
+/*	$OpenBSD: ntpd.c,v 1.105 2016/01/27 21:48:34 reyk Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -30,8 +30,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <err.h>
 
 #include "ntpd.h"
@@ -40,7 +42,7 @@ void		sighdlr(int);
 __dead void	usage(void);
 int		main(int, char *[]);
 int		check_child(pid_t, const char *);
-int		dispatch_imsg(struct ntpd_conf *);
+int		dispatch_imsg(struct ntpd_conf *, const char *, uid_t, gid_t);
 int		dispatch_imsg_ctl(struct ntpd_conf *);
 void		reset_adjtime(void);
 int		ntpd_adjtime(double);
@@ -53,6 +55,7 @@ const char     *ctl_lookup_option(char *, const char **);
 void		show_status_msg(struct imsg *);
 void		show_peer_msg(struct imsg *, int);
 void		show_sensor_msg(struct imsg *, int);
+void		update_time_sync_status(int);
 
 volatile sig_atomic_t	 quit = 0;
 volatile sig_atomic_t	 reconfig = 0;
@@ -60,6 +63,8 @@ volatile sig_atomic_t	 sigchld = 0;
 struct imsgbuf		*ibuf;
 int			 timeout = INFTIM;
 
+extern u_int		 constraint_cnt;
+
 const char		*showopt;
 
 static const char *ctl_showopt_list[] = {
@@ -111,6 +116,7 @@ usage(void)
 
 #define POLL_MAX		8
 #define PFD_PIPE		0
+#define PFD_MAX			1
 
 /* Saves a copy of argv for setproctitle emulation */
 #ifndef HAVE_SETPROCTITLE
@@ -123,13 +129,21 @@ int
 main(int argc, char *argv[])
 {
 	struct ntpd_conf	 lconf;
-	struct pollfd		 pfd[POLL_MAX];
+	struct pollfd		*pfd = NULL;
 	pid_t			 chld_pid = 0, pid;
 	const char		*conffile;
-	int			 fd_ctl, ch, nfds;
+	int			 fd_ctl, ch, nfds, i, j;
 	int			 pipe_chld[2];
-	struct passwd		*pw;
 	extern char		*__progname;
+	u_int			 pfd_elms = 0, new_cnt;
+	struct constraint	*cstr;
+	struct passwd		*pw;
+	const char		*pw_dir;
+	uid_t			pw_uid;
+	gid_t			pw_gid;
+	void			*newp;
+
+	__progname = get_progname(argv[0]);
 
 	if (strcmp(__progname, "ntpctl") == 0) {
 		ctl_main(argc, argv);
@@ -138,14 +152,9 @@ main(int argc, char *argv[])
 
 	conffile = CONFFILE;
 
-	bzero(&lconf, sizeof(lconf));
-
-	log_init(1);		/* log to stderr until daemonized */
-
-	__progname = get_progname(argv[0]);
+	memset(&lconf, 0, sizeof(lconf));
 
 #ifndef HAVE_SETPROCTITLE
-	int i;
 	/* Prepare for later setproctitle emulation */
 	saved_argv = calloc(argc + 1, sizeof(*saved_argv));
 	for (i = 0; i < argc; i++)
@@ -158,13 +167,13 @@ main(int argc, char *argv[])
 	while ((ch = getopt(argc, argv, "df:np:sSv")) != -1) {
 		switch (ch) {
 		case 'd':
-			lconf.debug = 1;
-			log_verbose(1);
+			lconf.debug = 2;
 			break;
 		case 'f':
 			conffile = optarg;
 			break;
 		case 'n':
+			lconf.debug = 2;
 			lconf.noaction = 1;
 			break;
 		case 'p':
@@ -177,7 +186,7 @@ main(int argc, char *argv[])
 			lconf.settime = 0;
 			break;
 		case 'v':
-			log_verbose(1);
+			lconf.verbose++;
 			break;
 		default:
 			usage();
@@ -185,6 +194,9 @@ main(int argc, char *argv[])
 		}
 	}
 
+	/* log to stderr until daemonized */
+	log_init(lconf.debug ? lconf.debug : 1, LOG_DAEMON);
+
 	argc -= optind;
 	argv += optind;
 	if (argc > 0)
@@ -204,12 +216,17 @@ main(int argc, char *argv[])
 	if ((pw = getpwnam(NTPD_USER)) == NULL)
 		errx(1, "unknown user %s", NTPD_USER);
 
+	pw_dir = strdup(pw->pw_dir);
+	pw_uid = pw->pw_uid;
+	pw_gid = pw->pw_gid;
+
 	if (setpriority(PRIO_PROCESS, 0, -20) == -1)
 		warn("can't set priority");
 
 	reset_adjtime();
 	if (!lconf.settime) {
-		log_init(lconf.debug);
+		log_init(lconf.debug, LOG_DAEMON);
+		log_verbose(lconf.verbose);
 		if (!lconf.debug) {
 			if (daemon(1, 0))
 				fatal("daemon");
@@ -230,7 +247,7 @@ main(int argc, char *argv[])
 	/* fork child process */
 	chld_pid = ntp_main(pipe_chld, fd_ctl, &lconf, pw);
 
-	setproctitle("[priv]");
+	log_procinit("[priv]");
 	readfreq();
 
 	signal(SIGTERM, sighdlr);
@@ -238,18 +255,49 @@ main(int argc, char *argv[])
 	signal(SIGHUP, sighdlr);
 
 	close(pipe_chld[1]);
+	constraint_purge();
 
 	if ((ibuf = malloc(sizeof(struct imsgbuf))) == NULL)
 		fatal(NULL);
 	imsg_init(ibuf, pipe_chld[0]);
 
+	constraint_cnt = 0;
+
+	/*
+	 * Constraint processes are forked with certificates in memory,
+	 * then privdrop into chroot before speaking to the outside world.
+	 */
+	if (pledge("stdio rpath inet settime proc id", NULL) == -1)
+		err(1, "pledge");
+
 	while (quit == 0) {
+		new_cnt = PFD_MAX + constraint_cnt;
+		if (new_cnt > pfd_elms) {
+			if ((newp = reallocarray(pfd, new_cnt,
+			    sizeof(*pfd))) == NULL) {
+				/* panic for now */
+				log_warn("could not resize pfd from %u -> "
+				    "%u entries", pfd_elms, new_cnt);
+				fatalx("exiting");
+			}
+			pfd = newp;
+			pfd_elms = new_cnt;
+		}
+
+		memset(pfd, 0, sizeof(*pfd) * pfd_elms);
 		pfd[PFD_PIPE].fd = ibuf->fd;
 		pfd[PFD_PIPE].events = POLLIN;
 		if (ibuf->w.queued)
 			pfd[PFD_PIPE].events |= POLLOUT;
 
-		if ((nfds = poll(pfd, 1, timeout)) == -1)
+		i = PFD_MAX;
+		TAILQ_FOREACH(cstr, &conf->constraints, entry) {
+			pfd[i].fd = cstr->fd;
+			pfd[i].events = POLLIN;
+			i++;
+		}
+
+		if ((nfds = poll(pfd, i, timeout)) == -1)
 			if (errno != EINTR) {
 				log_warn("poll error");
 				quit = 1;
@@ -258,7 +306,8 @@ main(int argc, char *argv[])
 		if (nfds == 0 && lconf.settime) {
 			lconf.settime = 0;
 			timeout = INFTIM;
-			log_init(lconf.debug);
+			log_init(lconf.debug, LOG_DAEMON);
+			log_verbose(lconf.verbose);
 			log_warnx("no reply received in time, skipping initial "
 			    "time setting");
 			if (!lconf.debug) {
@@ -276,10 +325,14 @@ main(int argc, char *argv[])
 
 		if (nfds > 0 && pfd[PFD_PIPE].revents & POLLIN) {
 			nfds--;
-			if (dispatch_imsg(&lconf) == -1)
+			if (dispatch_imsg(&lconf, pw_dir, pw_uid, pw_gid) == -1)
 				quit = 1;
 		}
 
+		for (j = PFD_MAX; nfds > 0 && j < i; j++) {
+			nfds -= priv_constraint_dispatch(&pfd[j]);
+		}
+
 		if (sigchld) {
 			if (check_child(chld_pid, "child")) {
 				quit = 1;
@@ -310,36 +363,46 @@ main(int argc, char *argv[])
 }
 
 int
-check_child(pid_t pid, const char *pname)
+check_child(pid_t chld_pid, const char *pname)
 {
 	int	 status, sig;
 	char	*signame;
+	pid_t	 pid;
 
-	if (waitpid(pid, &status, WNOHANG) > 0) {
-		if (WIFEXITED(status)) {
-			log_warnx("Lost child: %s exited", pname);
-			return (1);
+	do {
+		pid = waitpid(WAIT_ANY, &status, WNOHANG);
+		if (pid <= 0) {
+			continue;
+		} else if (pid == chld_pid) {
+			if (WIFEXITED(status)) {
+				log_warnx("Lost child: %s exited", pname);
+				return (1);
+			}
+			if (WIFSIGNALED(status)) {
+				sig = WTERMSIG(status);
+				signame = strsignal(sig) ?
+				    strsignal(sig) : "unknown";
+				log_warnx("Lost child: %s terminated; "
+				    "signal %d (%s)", pname, sig, signame);
+				return (1);
+			}
+		} else {
+			priv_constraint_check_child(pid, status);
 		}
-		if (WIFSIGNALED(status)) {
-			sig = WTERMSIG(status);
-			signame = strsignal(sig) ? strsignal(sig) : "unknown";
-			log_warnx("Lost child: %s terminated; signal %d (%s)",
-			    pname, sig, signame);
-			return (1);
-		}
-	}
+	} while (pid > 0 || (pid == -1 && errno == EINTR));
 
 	return (0);
 }
 
 int
-dispatch_imsg(struct ntpd_conf *lconf)
+dispatch_imsg(struct ntpd_conf *lconf, const char *pw_dir,
+    uid_t pw_uid, gid_t pw_gid)
 {
 	struct imsg		 imsg;
 	int			 n;
 	double			 d;
 
-	if ((n = imsg_read(ibuf)) == -1)
+	if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
 		return (-1);
 
 	if (n == 0) {	/* connection closed */
@@ -374,7 +437,8 @@ dispatch_imsg(struct ntpd_conf *lconf)
 				fatalx("invalid IMSG_SETTIME received");
 			if (!lconf->settime)
 				break;
-			log_init(lconf->debug);
+			log_init(lconf->debug, LOG_DAEMON);
+			log_verbose(lconf->verbose);
 			memcpy(&d, imsg.data, sizeof(d));
 			ntpd_settime(d);
 			/* daemonize now */
@@ -386,6 +450,14 @@ dispatch_imsg(struct ntpd_conf *lconf)
 			lconf->settime = 0;
 			timeout = INFTIM;
 			break;
+		case IMSG_CONSTRAINT_QUERY:
+			priv_constraint_msg(imsg.hdr.peerid,
+			    imsg.data, imsg.hdr.len - IMSG_HEADER_SIZE,
+			    pw_dir, pw_uid, pw_gid);
+			break;
+		case IMSG_CONSTRAINT_KILL:
+			priv_constraint_kill(imsg.hdr.peerid);
+			break;
 		default:
 			break;
 		}
@@ -423,6 +495,7 @@ ntpd_adjtime(double d)
 	else if (!firstadj && olddelta.tv_sec == 0 && olddelta.tv_usec == 0)
 		synced = 1;
 	firstadj = 0;
+	update_time_sync_status(synced);
 	return (synced);
 }
 
@@ -487,61 +560,62 @@ ntpd_settime(double d)
 	log_info("set local clock to %s (offset %fs)", buf, d);
 }
 
+static FILE *freqfp;
+
 void
 readfreq(void)
 {
-	FILE *fp;
 	int64_t current;
+	int fd;
 	double d;
 
-	fp = fopen(DRIFTFILE, "r");
-	if (fp == NULL) {
-		/* if the drift file has been deleted by the user, reset */
+	fd = open(DRIFTFILE, O_RDWR);
+	if (fd == -1) {
+		log_warnx("creating new %s", DRIFTFILE);
 		current = 0;
 		if (adjfreq(&current, NULL) == -1)
 			log_warn("adjfreq reset failed");
+		freqfp = fopen(DRIFTFILE, "w");
 		return;
 	}
 
+	freqfp = fdopen(fd, "r+");
+
 	/* if we're adjusting frequency already, don't override */
 	if (adjfreq(NULL, &current) == -1)
 		log_warn("adjfreq failed");
-	else if (current == 0) {
-		if (fscanf(fp, "%lf", &d) == 1) {
+	else if (current == 0 && freqfp) {
+		if (fscanf(freqfp, "%lf", &d) == 1) {
 			d /= 1e6;	/* scale from ppm */
 			ntpd_adjfreq(d, 0);
 		} else
-			log_warnx("can't read %s", DRIFTFILE);
+			log_warnx("%s is empty", DRIFTFILE);
 	}
-	fclose(fp);
 }
 
 int
 writefreq(double d)
 {
 	int r;
-	FILE *fp;
 	static int warnonce = 1;
+        int off;
 
-	fp = fopen(DRIFTFILE, "w");
-	if (fp == NULL) {
-		if (warnonce) {
-			log_warn("can't open %s", DRIFTFILE);
-			warnonce = 0;
-		}
+	if (freqfp == NULL)
 		return 0;
-	}
-
-	fprintf(fp, "%.3f\n", d * 1e6);		/* scale to ppm */
-	r = ferror(fp);
-	if (fclose(fp) != 0 || r != 0) {
+	rewind(freqfp);
+	r = fprintf(freqfp, "%.3f\n", d * 1e6);	/* scale to ppm */
+	if (r < 0 || fflush(freqfp) != 0) {
 		if (warnonce) {
 			log_warnx("can't write %s", DRIFTFILE);
 			warnonce = 0;
 		}
-		unlink(DRIFTFILE);
+		clearerr(freqfp);
 		return 0;
 	}
+        off = ftello(freqfp);
+        if (off == -1 || ftruncate(fileno(freqfp), off) == -1)
+                log_warnx("can't truncate %s", DRIFTFILE);
+        fsync(fileno(freqfp));
 	return 1;
 }
 
@@ -600,7 +674,7 @@ ctl_main(int argc, char *argv[])
 	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
 		err(1, "ntpctl: socket");
 
-	bzero(&sa, sizeof(sa));
+	memset(&sa, 0, sizeof(sa));
 	sa.sun_family = AF_UNIX;
 	if (strlcpy(sa.sun_path, sockname, sizeof(sa.sun_path)) >=
 	    sizeof(sa.sun_path))
@@ -608,6 +682,9 @@ ctl_main(int argc, char *argv[])
 	if (connect(fd, (struct sockaddr *)&sa, sizeof(sa)) == -1)
 		err(1, "connect: %s", sockname);
 
+	if (pledge("stdio", NULL) == -1)
+		err(1, "pledge");
+
 	if ((ibuf_ctl = malloc(sizeof(struct imsgbuf))) == NULL)
 		err(1, NULL);
 	imsg_init(ibuf_ctl, fd);
@@ -640,7 +717,7 @@ ctl_main(int argc, char *argv[])
 
 	done = 0;
 	while (!done) {
-		if ((n = imsg_read(ibuf_ctl)) == -1)
+		if ((n = imsg_read(ibuf_ctl)) == -1 && errno != EAGAIN)
 			err(1, "ibuf_ctl: imsg_read error");
 		if (n == 0)
 			errx(1, "ntpctl: pipe closed");
